import {
  ConflictException,
  Injectable,
  NotFoundException,
} from '@nestjs/common';
import { InjectRepository } from '@nestjs/typeorm';
import { Not, Repository } from 'typeorm';
import { Infraccion } from './entity/infraccion.entity';
import { InstitucionInfraccion } from './entity/institucion-infraccion.entity';
import { Operador } from '../operador/entity/operador.entity';
import { InstitucionService } from '../institucion/institucion.service';
import { Institucion } from 'src/institucion/entity/institucion.entity';

@Injectable()
export class InstitucionInfraccionService {
  constructor(
    @InjectRepository(Infraccion)
    private infraccionRepository: Repository<Infraccion>,
    @InjectRepository(InstitucionInfraccion)
    private institucionInfraccionRepository: Repository<InstitucionInfraccion>,
    private institucionService: InstitucionService,
  ) {}

  create(infraccion: string) {
    return this.infraccionRepository
      .findOne({ where: { infraccion } })
      .then((existeInfraccion) => {
        if (existeInfraccion)
          throw new ConflictException('Ya existe esta infracción.');
        return this.infraccionRepository.save(
          this.infraccionRepository.create({ infraccion }),
        );
      })
      .then(async (infraccion) => {
        const instituciones =
          await this.institucionService.findInfoInstitucionAll();

        for (let i = 0; i < instituciones.length; i++)
          await this.institucionInfraccionRepository.save(
            this.institucionInfraccionRepository.create({
              infraccion,
              institucion: instituciones[i],
            }),
          );
        return { message: 'Se creó correctamente una nueva infracción.' };
      });
  }

  findAll() {
    return this.infraccionRepository.find({
      where: { id_infraccion: Not(1) },
      order: { infraccion: 'ASC' },
    });
  }

  findAllByIdInstitucion(id_institucion: number) {
    return this.institucionService
      .findInfoInstitucionById(id_institucion)
      .then((institucion) =>
        this.institucionInfraccionRepository
          .createQueryBuilder('ii')
          .innerJoinAndSelect('ii.infraccion', 'inf')
          .innerJoinAndSelect('ii.institucion', 'i')
          .where(
            'i.id_institucion = :id_institucion && inf.id_infraccion != 1',
            { id_institucion: institucion.id_institucion },
          )
          .orderBy('inf.infraccion')
          .getMany(),
      );
  }

  findById(id_institucion_infraccion: number) {
    return this.institucionInfraccionRepository
      .findOne({
        join: {
          alias: 'ii',
          innerJoinAndSelect: { i: 'ii.institucion', in: 'ii.infraccion' },
        },
        where: { id_institucion_infraccion },
      })
      .then((institucionInfraccion) => {
        if (!institucionInfraccion)
          throw new NotFoundException(
            'No existe este id institución infracción.',
          );
        return institucionInfraccion;
      });
  }

  findSinInfraccion(institucion: Institucion) {
    return this.institucionInfraccionRepository
      .createQueryBuilder('ii')
      .innerJoinAndSelect('ii.infraccion', 'inf')
      .innerJoinAndSelect('ii.institucion', 'i')
      .where('i.id_institucion = :id_institucion && inf.id_infraccion = 1', {
        id_institucion: institucion.id_institucion,
      })
      .getOne();
  }

  update(admin: Operador, attrs: Partial<InstitucionInfraccion>) {
    return this.findById(attrs.id_institucion_infraccion)
      .then((institucionInfraccion) => {
        if (
          admin.institucion.id_institucion !=
          institucionInfraccion.institucion.id_institucion
        )
          throw new ConflictException(
            'No puedes actualizar la información de esta infraccion porque no le pertenece a tu institución.',
          );
        Object.assign(institucionInfraccion, attrs);
        return this.institucionInfraccionRepository.save(institucionInfraccion);
      })
      .then((_) => ({
        message: 'Se guardaron los cambios correctamente.',
      }));
  }
}
