import {
  ConflictException,
  Injectable,
  NotFoundException,
} from '@nestjs/common';
import { InjectRepository } from '@nestjs/typeorm';
import { Between, FindOptionsWhere, Like, Not, Repository } from 'typeorm';
import { Institucion } from '../institucion/entity/institucion.entity';
import { Operador } from './entity/operador.entity';
import { FullInformacionOperadorView } from './entity/views/full-informacion-operador.view';
import { BcryptService } from '../bcrypt/bcrypt.service';
import { InstitucionService } from '../institucion/institucion.service';
import { NodemailerService } from '../nodemailer/nodemailer.service';
import { TipoUsuarioService } from '../tipo-usuario/tipo-usuario.service';

@Injectable()
export class OperadorService {
  constructor(
    @InjectRepository(Operador) private repository: Repository<Operador>,
    @InjectRepository(FullInformacionOperadorView)
    private fullInformacionOperadorView: Repository<FullInformacionOperadorView>,
    private bcryptService: BcryptService,
    private institucionService: InstitucionService,
    private nodemailerService: NodemailerService,
    private tipoUsuarioService: TipoUsuarioService,
  ) {}

  async create(
    id_institucion: number | Institucion,
    id_tipo_usuario: number,
    operador: string,
    nombre: string,
    correo: string,
    password?: string,
  ) {
    const institucion =
      typeof id_institucion === 'number'
        ? await this.institucionService.findInfoInstitucionById(id_institucion)
        : id_institucion;
    const tipoUsuario = await this.tipoUsuarioService.findById(id_tipo_usuario);

    // Solo se puede crear admins y operadores
    if (tipoUsuario.id_tipo_usuario < 3 || tipoUsuario.id_tipo_usuario > 4)
      throw new ConflictException(
        'No se puede asignar un tipo de usuario distinto a admin y operador',
      );
    // Ver si existe un admin con ese usaurio
    return this.findInfoOperadorAdmin(operador)
      .then((existeAdmin) => {
        // Error si lo hay
        if (existeAdmin)
          throw new ConflictException(
            'Ya existe un admin con ese nombre, intenta de nuevo con otro.',
          );
        // Ver que haya un operador en esta institución con este usuario
        return this.findInfoOperador(institucion, operador);
      })
      .then(async (existeOperador) => {
        // Error si lo hay
        if (existeOperador)
          throw new ConflictException(
            'Ya existe un operador en esta institución con ese nombre, intenta de nuevo con otro.',
          );
        // Crear password si no se mandó
        if (!password) password = this.bcryptService.generarPassword();
        // Crear registro
        return this.repository.save(
          this.repository.create({
            correo,
            institucion,
            nombre,
            operador,
            password: this.bcryptService.encriptar(password),
            tipoUsuario,
          }),
        );
      })
      .then((operador) =>
        // Correo con credenciales
        this.nodemailerService.sendEmail({
          email: operador.correo,
          subject: 'Credenciales Pc Puma',
          html:
            tipoUsuario.id_tipo_usuario === 3
              ? this.nodemailerService.correoPasswordAdmin(
                  operador.operador,
                  password,
                )
              : this.nodemailerService.correoPasswordOperador(
                  operador.operador,
                  password,
                ),
        }),
      )
      .then((_) => ({
        message: `Se creó correctamente un nuevo ${tipoUsuario.tipo_usuario.toLowerCase()}.`,
      }));
  }

  findAllByInstitucion(operador: Operador) {
    return this.repository.find({
      select: ['id_operador', 'operador'],
      where: { institucion: operador.institucion },
    });
  }

  async findFullInfoOperadorAll(filtros: {
    pagina: string;
    id_institucion?: string;
    id_tipo_usuario?: string;
    operador?: string;
  }) {
    const institucion = filtros.id_institucion
      ? await this.institucionService.findInfoInstitucionById(
          parseInt(filtros.id_institucion),
        )
      : null;
    const tipoUsuario = filtros.id_tipo_usuario
      ? await this.tipoUsuarioService.findById(
          parseInt(filtros.id_tipo_usuario),
        )
      : null;
    const busqueda: FindOptionsWhere<FullInformacionOperadorView> = {};

    if (filtros.operador) busqueda.operador = Like(`${filtros.operador}%`);
    if (institucion) busqueda.id_institucion = institucion.id_institucion;
    if (tipoUsuario) busqueda.id_tipo_usuario = tipoUsuario.id_tipo_usuario;
    return this.fullInformacionOperadorView
      .findAndCount({
        where: [
          busqueda,
          { id_tipo_usuario: Not(1) },
          { id_tipo_usuario: Not(2) },
        ],
        take: 25,
        skip: (parseInt(filtros.pagina) - 1) * 25,
      })
      .then((infoOperadores) => {
        const operadores: Operador[] = [];

        for (let i = 0; i < infoOperadores[0].length; i++)
          operadores.push(
            this.repository.create({
              id_operador: infoOperadores[0][i].id_operador,
              activo: infoOperadores[0][i].activo,
              nombre: infoOperadores[0][i].nombre,
              operador: infoOperadores[0][i].operador,
              institucion: {
                id_institucion: infoOperadores[0][i].id_operador,
                institucion: infoOperadores[0][i].institucion,
              },
              tipoUsuario: {
                id_tipo_usuario: infoOperadores[0][i].id_tipo_usuario,
                tipo_usuario: infoOperadores[0][i].tipo_usuario,
              },
            }),
          );
        return [operadores];
      });
  }

  findInfoOperador(institucion: Institucion, operador: string) {
    return this.repository.findOne({
      select: {
        id_operador: true,
        activo: true,
        nombre: true,
        password: true,
        institucion: { id_institucion: true },
        tipoUsuario: { id_tipo_usuario: true },
      },
      where: { operador, institucion },
    });
  }

  findInfoOperadorAdmin(admin: string) {
    return this.repository.findOne({
      select: {
        id_operador: true,
        activo: true,
        nombre: true,
        password: true,
        institucion: { id_institucion: true },
        tipoUsuario: { id_tipo_usuario: true },
      },
      where: {
        operador: admin,
        tipoUsuario: { id_tipo_usuario: Between(2, 3) },
      },
    });
  }

  findInfoOperadorById(id_operador: number) {
    return this.repository
      .findOne({
        select: {
          id_operador: true,
          activo: true,
          correo : true,
          institucion: { id_institucion: true },
          tipoUsuario: { id_tipo_usuario: true },
        },
        where: { id_operador },
      })
      .then((operador) => {
        if (!operador)
          throw new NotFoundException('No existe este id operador.');
        return operador;
      });
  }

  passwordReset(admin: Operador, id_operador: number, password?: string) {
    return this.findInfoOperadorById(id_operador)
      .then((operador) => {
        this.validarUpdate(admin, operador);
        // Si no se manda una password se genera una
        if (!password) password = this.bcryptService.generarPassword();
        // Encriptación
        operador.password = this.bcryptService.encriptar(password);
        // Guardar
        return this.repository.save(operador);
      })
      .then((operador) =>
        // Reenviar correo
        this.nodemailerService.sendEmail({
          email: operador.correo,
          subject: 'Credenciales PC Puma',
          html:
            operador.tipoUsuario.id_tipo_usuario === 3
              ? this.nodemailerService.correoPasswordAdmin(
                  operador.operador,
                  password,
                )
              : this.nodemailerService.correoPasswordOperador(
                  operador.operador,
                  password,
                ),
        }),
      )
      .then((_) => ({ message: 'Se cambió correctamente la constraseña.' }));
  }

  update(admin: Operador, attrs: Partial<Operador>) {
    return this.findInfoOperadorById(attrs.id_operador)
      .then((operador) => {
        this.validarUpdate(admin, operador);
        // Paso los valores enviados del front a mi objeto
        Object.assign(operador, attrs);
        // Guardar
        return this.repository.save(operador);
      })
      .then((_) => ({ message: 'Se guardaron los cambios correctamente.' }));
  }

  validarUpdate(admin: Operador, operador: Operador) {
    // Valido que el super admin solo puedar modificar admins
    if (
      admin.tipoUsuario.id_tipo_usuario === 2 &&
      operador.tipoUsuario.id_tipo_usuario != 3
    )
      throw new ConflictException(
        `El super admin solo puede modificar la infomración de los admins.`,
      );
    if (admin.tipoUsuario.id_tipo_usuario === 3) {
      // Valido admin no pueda modificar admins
      if (operador.tipoUsuario.id_tipo_usuario === 3)
        throw new ConflictException(
          `No puedes actualizar la información de otro admin.`,
        );
      // Valido que admin solo pueda modificar operadores de su institución
      if (
        admin.institucion.id_institucion != operador.institucion.id_institucion
      )
        throw new ConflictException(
          `No puedes actualizar la información de este operador porque no pertenece a tu institución.`,
        );
    }
  }
}
